// {$MODE DELPHI}
{$MODE OBJFPC}
{$MODESWITCH NESTEDCOMMENTS+}

unit EncPRNG;

(*------------------------------*)
{
v1.0	Derived from Encrypt_P.pas v2.10
v1.1	1. Remove redundant vars
		2. fixed LCG_table size of 6.
		3. extra moduplier checks
v1.2	Lots of changes.
v1.2X	Detect repeating RNGs and increment seed.
v3.0	Repeating RNGs very big fux
v3.1	Lots of fixes/changes derived from 3.0X
v3.2	1.	LCG modulus/multiplier switched to make
			"most significant" LCG more random.
		2.	Reduce no of LCG's back to 6 from 7.
		3.	Scale MILLIPASS_BIGINT better to
			fit exactly 6 LCG's.
}

interface

uses	sysutils
,		strutils
,		Interfaces
,		Multi_Int
;

{$IFDEF CPU64}
	{$define 64bit}
{$ELSE}
  	{$define 32bit}
{$ENDIF}

const
VERSION				= '3.2';
WHO_I_AM			= 'Marks PRNG';
WHO_I_AM_VERSION	= WHO_I_AM + ' v' + VERSION;

const
	INT8_MAXINT		= 127;
	INT8_MAXINT_1	= 128;
	INT8U_MAXINT	= 255;
	INT8U_MAXINT_1	= 256;
	INT16_MAXINT	= 32767;
	INT16_MAXINT_1	= 32768;
	INT16U_MAXINT	= 65535;
	INT16U_MAXINT_1 = 65536;
	INT32_MAXINT	= 2147483647;
	INT32_MAXINT_1	= 2147483648;
	INT32U_MAXINT	= 4294967295;
	INT32U_MAXINT_1	= 4294967296;
	INT64_MAXINT	= 9223372036854775807;
	INT64_MAXINT_1	= 9223372036854775808;
	INT64U_MAXINT	= 18446744073709551615;
	INT64U_MAXINT_1 = 18446744073709551616;

{
	ADD_TRAILING_COUNT	= TRUE;
	WITH_DELIMITERS		= TRUE;
	NO_DELIMITERS		= FALSE;
	C64_LINE_LENGTH		= 40;
}

type
	int8u		= byte;
	int8		= shortint;
	int16		= smallint;
	int16u		= word;
	int32		= longint;
	int32u		= longword;
	int64u		= QWord;
(*	int64		= int64 *)

type

	big_int_type	= Multi_Int_XV;
	int_type		= int64;
	string_type		= ansistring;
	T_OP			= (T_OP_NULL,T_OP_WRITE_RNG,T_OP_WRITE_PRIMES,T_OP_WRITE_INDEXES);
	T_TEST_TYPE		= (T_TEST_NULL,T_WRITE_RNG_TEXT,T_WRITE_RNG_BIN,T_WRITE_MILLI,T_OP_WRITE_SEEDS);
	T_ERROR_CODE	= (T_ERR_OK, T_ERR_HEX, T_ERR_C64, T_ERR_INTERNAL, T_ERR_PARAM, T_ERR_FILE, T_ERR_HEADER, T_ERR_MODULUS);

var
	OPERATION			:T_OP;
	TEST_MODE			:BOOLEAN;
	TEST_TYPE			:T_TEST_TYPE;
	NUMBER_PARAM		:int_type = 1;
	WRITE_RNG_BITSIZE	:int_type;
	RANDMODNUM			:int_type = INT8U_MAXINT_1;
	MILLISECS_PARAM		:int_type;

	ERROR_CODE		:T_ERROR_CODE;
	ERROR_MESS		:string_type;

procedure	initialise_Encrypt_P;
procedure 	init_rand_modupliers;
procedure	GET_TIMESTAMP_FROM_SYSTEM_TIME;
procedure	init_MILLIPASS_BIGINT;
procedure 	read_params;
procedure	write_rng;
procedure	write_primes;

implementation

const

TWO_POWER_27		= 134217728;		// this is used to scale days_since when combining with millisecs daysince
TWO_POWER_40		= 1099511627776;	// this is used to scale days_since when combining with microsecs daysince

MIN_PRIME			= 420000;
MAX_PRIME			= 100000000;
REPEAT_MAX			= 15;

var
TimeStamp		:TTimeStamp;
RANDNUM_PARAM	:int_type = -1;
INTERNAL_RANDOM	:int_type;
MILLISECS		:int_type;
MICROSECS		:int_type;
DAYSINCE    	:int_type;
MICRODAYS,
PICRODAYS   	:int64u;
LCG_table		:array of
					RECORD
					MULTIPLIER	:int_type;
					CONSTANT	:int_type;
					MODULUS		:int_type;
					SEED		:int_type;
					SHIFT		:int_type;
					END;
LCG_current_seed:array of int_type;
LCG_table_i,
LCG_table_max	:int_type;
SEEDS_EXHAUSTED	:boolean;

REPEAT_TABLE_FULL	:boolean;
REPEAT_TABLE		:ARRAY[1..REPEAT_MAX] OF int_type;
REPEAT_COUNT		:int_type;
T_COUNT				:int_type;
RR_COUNT			:int_type;
TT_COUNT			:int_type;

DIGIT				:array [0..3] of int_type = (1,3,7,9);

MILLIPASS_BIGINT,
MILLIPASS_BIGINT_TEMP	:big_int_type;

(*------------------------------*)
function DEFAULT_RAND_FROM_TIMER:uint32;
var
// COUNTAVER,
COUNTER		:uint32;
DATETIME,
DATETIMENEXT:TTimeStamp;
begin

DATETIME:= DateTimeToTimeStamp(Now);
COUNTER:= 0;
repeat
	Inc(COUNTER);
	DATETIMENEXT:= DateTimeToTimeStamp(Now);
until	((DATETIMENEXT.Time MOD 10) = 0)
and		(DATETIMENEXT.Time <> DATETIME.Time)
and		(COUNTER > 1)
;
COUNTER:= (COUNTER MOD 10000);
Result:= COUNTER;
end;


(*------------------------------*)
procedure GET_TIMESTAMP_FROM_SYSTEM_TIME;
begin
TimeStamp:= DateTimeToTimeStamp(Now);
end;


(*--------------V2--------------*)
procedure GET_MICRODAYS_FROM_TIMESTAMP;
var	D	:int_type;
begin
MILLISECS:= TimeStamp.time;
DAYSINCE:= TimeStamp.Date;

if (MILLISECS_PARAM > 0) then
    DAYSINCE:= MILLISECS_PARAM;

MICROSECS:= ((MILLISECS * 10000) + INTERNAL_RANDOM);
MICRODAYS:= ((DAYSINCE * TWO_POWER_40) + MICROSECS);

D:= (MICRODAYS and 3);
PICRODAYS:= (((MICRODAYS >> 2) * 10) + DIGIT[D]);

end;


(*--------------v3B--------------*)
procedure init_rand_modupliers;
const
	PRIMES_PER_LCG		= 2;
	NUM_PRIMES_DIGITS	= 8;
	NUM_SHIFT_DIGITS	= 1;
var
	MPX		:string_type;
	R,I,T,
	DUP_MODUPLIER_COUNT,
    MODUPLIER_RANGE,
	PTI_INT	:int_type;
	PTI		:string_type;

(*------*)
function M_VALUE(PR:int_type):int_type;
	var
	D,R,V	:int_type;
	begin
	R:= (PR + MIN_PRIME);
	D:= (R and 3);
	V:= ((R >> 2) * 10);
	V:= (V + DIGIT[D]);
	Result:= V;
	end;

(*------*)
procedure check_duplicate_moduplier;
begin
I:= 0;
while (I < LCG_table_i) do
	begin
	if	(	(LCG_table[LCG_table_i].MODULUS = LCG_table[I].MODULUS)
	//	or	(LCG_table[LCG_table_i].MULTIPLIER = LCG_table[I].MULTIPLIER)
		or	(LCG_table[I].MULTIPLIER = LCG_table[I].MODULUS)
		)
	then Inc(DUP_MODUPLIER_COUNT);
	Inc(I);
	end;
end;

(*------*)
procedure check_moduplier_range;
begin
I:= 0;
MODUPLIER_RANGE:= 0;
while (I < LCG_table_i) do
	begin
	T:= ABS(LCG_table[0].MODULUS - LCG_table[I].MODULUS);
	if (T > MODUPLIER_RANGE)
	then MODUPLIER_RANGE:= T;
	T:= ABS(LCG_table[0].MULTIPLIER - LCG_table[I].MULTIPLIER);
	if (T > MODUPLIER_RANGE)
	then MODUPLIER_RANGE:= T;
	T:= ABS(LCG_table[0].MULTIPLIER - LCG_table[I].MODULUS);
	if (T > MODUPLIER_RANGE)
	then MODUPLIER_RANGE:= T;
	Inc(I);
	end;
end;

(*------*)
begin
DUP_MODUPLIER_COUNT:= 0;

MPX:= MILLIPASS_BIGINT.tostr;
LCG_table_max:= ( length(MPX) div ((NUM_PRIMES_DIGITS * PRIMES_PER_LCG) + NUM_SHIFT_DIGITS) );

if (OPERATION = T_OP_WRITE_INDEXES)
or (OPERATION = T_OP_WRITE_PRIMES)
then
	begin
	writeln('LCG_table_max = ',LCG_table_max);
	end;

setlength(LCG_table,LCG_table_max);

LCG_table_i:= 0;
while (LCG_table_i < LCG_table_max) do
	begin

    PTI:= (midstr(MPX,(length(MPX)-NUM_SHIFT_DIGITS+1),NUM_SHIFT_DIGITS));
	MPX:= midstr(MPX,1,(length(MPX)-NUM_SHIFT_DIGITS));
	PTI_INT:= strtoint64(PTI);
	R:= PTI_INT;
    LCG_table[LCG_table_i].CONSTANT:= 1;
	LCG_table[LCG_table_i].SEED:= 1;
	LCG_table[LCG_table_i].SHIFT:= ((R and 3) + 1);

	if (OPERATION = T_OP_WRITE_INDEXES) then write('PTI =');

    PTI:= (midstr(MPX,(length(MPX)-NUM_PRIMES_DIGITS+1),NUM_PRIMES_DIGITS));
	MPX:= midstr(MPX,1,(length(MPX)-NUM_PRIMES_DIGITS));
	PTI_INT:= strtoint64(PTI);
	if (OPERATION = T_OP_WRITE_INDEXES) then write(' ',PTI);
    LCG_table[LCG_table_i].MODULUS:= M_VALUE(PTI_INT);

    PTI:= (midstr(MPX,(length(MPX)-NUM_PRIMES_DIGITS+1),NUM_PRIMES_DIGITS));
	MPX:= midstr(MPX,1,(length(MPX)-NUM_PRIMES_DIGITS));
	PTI_INT:= strtoint64(PTI);
	if (OPERATION = T_OP_WRITE_INDEXES) then write(' ',PTI);
    LCG_table[LCG_table_i].MULTIPLIER:= M_VALUE(PTI_INT);

	if (OPERATION = T_OP_WRITE_INDEXES) then writeln;

	if (OPERATION = T_OP_WRITE_INDEXES) then
		begin
        write(LCG_table[LCG_table_i].MULTIPLIER,' ');
        write(LCG_table[LCG_table_i].MODULUS,' ');
		write(LCG_table[LCG_table_i].SHIFT,' ');
		writeln;
		writeln;
		end;

	check_duplicate_moduplier;

    inc(LCG_table_i);
	end;
check_moduplier_range;

{
if (OPERATION = T_OP_WRITE_PRIMES) then
	begin
	writeln('SAME MODULUS COUNT = ',DUP_MODUPLIER_COUNT);
	end;
}
if	(	(DUP_MODUPLIER_COUNT >= 1)
	OR	(MODUPLIER_RANGE < 1000)
	)
and	(ERROR_CODE = T_ERR_OK) then
    begin
	ERROR_CODE:= T_ERR_MODULUS;
	ERROR_MESS:= 'Unable to create sufficient randomness to seed the RNG - Please try again';
    end;
end;


(*------------------------------*)
procedure init_MILLIPASS_BIGINT;
// var
// b,c,e		:int_type;
// p			:int8u;
// LONGPASS	:string_type;

begin
MILLIPASS_BIGINT:= PICRODAYS;
MILLIPASS_BIGINT:= ((MILLIPASS_BIGINT ** 5) * (MILLIPASS_BIGINT DIV 100000000));

if MILLIPASS_BIGINT.Overflow then
	begin
	ERROR_CODE:= T_ERR_INTERNAL;
	ERROR_MESS:= 'MILLIPASS_BIGINT overflow!';
    exit;
	end;

if (OPERATION = T_OP_WRITE_INDEXES)
then
	begin
	writeln('MILLIPASS_BIGINT = ',MILLIPASS_BIGINT.tostr);
	writeln('DAYSINCE = ',DAYSINCE);
	writeln('MICROSECS = ',MICROSECS);
	writeln('MICRODAYS = ',MICRODAYS);
	writeln('PICRODAYS = ',PICRODAYS);
	end;
end;


(*------------------------------*)
procedure init_pass_rand;
var	retry:int_type;
begin
REPEAT_TABLE_FULL:= FALSE;
INTERNAL_RANDOM:= DEFAULT_RAND_FROM_TIMER;
GET_TIMESTAMP_FROM_SYSTEM_TIME;
GET_MICRODAYS_FROM_TIMESTAMP;
init_MILLIPASS_BIGINT;
init_rand_modupliers;

retry:= 0;
while	(ERROR_CODE = T_ERR_MODULUS)
and		(retry < 5)
do
	begin
	Inc(retry);
	if (OPERATION = T_OP_WRITE_INDEXES) then
		writeln('Duplicate modupliers');
	ERROR_CODE:= T_ERR_OK;

	// sleep(1);
	INTERNAL_RANDOM:= DEFAULT_RAND_FROM_TIMER;
	GET_TIMESTAMP_FROM_SYSTEM_TIME;
	GET_MICRODAYS_FROM_TIMESTAMP;
	init_MILLIPASS_BIGINT;
	init_rand_modupliers;

	if MILLIPASS_BIGINT.Overflow then
		begin
		ERROR_CODE:= T_ERR_INTERNAL;
		ERROR_MESS:= 'Internal error: MILLIPASS_BIGINT overflow!';
		end;

	if (OPERATION = T_OP_WRITE_INDEXES)
	then
		begin
		writeln('MILLIPASS_BIGINT = ',MILLIPASS_BIGINT.tostr);
		writeln('DAYSINCE = ',DAYSINCE);
		writeln('MICRODAYS = ',MICRODAYS);
		end;
	end;
end;


(*-----------------------*)
function rand_repeat_test(rand_var:int_type):boolean;
var	RETRO_MATCH:boolean;
begin
Result:= FALSE;

if REPEAT_TABLE_FULL then
	begin
    if  ( REPEAT_TABLE[REPEAT_COUNT] = rand_var ) then
        begin
		if   (REPEAT_COUNT = REPEAT_MAX) then
	        begin
			if	(TEST_TYPE = T_WRITE_RNG_TEXT) then
	        	writeln('Repeats');
	        Result:= TRUE;
			REPEAT_COUNT:= 1;
	        end
		else
			begin
	        Inc(REPEAT_COUNT);
			end;
		end
	else
    	if (REPEAT_COUNT > 1) then
			begin
			RETRO_MATCH:= FALSE;
			T_COUNT:= 2;
	        while (RETRO_MATCH = FALSE)	do
				begin
				TT_COUNT:= T_COUNT;
				RR_COUNT:= 1;
				while	(TT_COUNT < REPEAT_COUNT)
				and		(REPEAT_TABLE[TT_COUNT] = REPEAT_TABLE[RR_COUNT])
				do	begin
					Inc(TT_COUNT); Inc(RR_COUNT);
					end;
				if (TT_COUNT < REPEAT_COUNT) then
					Inc(T_COUNT)
				else
					begin
	            	RETRO_MATCH:= TRUE;
		            if	(REPEAT_TABLE[RR_COUNT] = rand_var) then
	                    REPEAT_COUNT:= (RR_COUNT + 1)
					else
						REPEAT_COUNT:= 1;
					end;
	            end;
			end;
	end
else
	begin
	if (REPEAT_COUNT = 0) then Inc(REPEAT_COUNT);
    REPEAT_TABLE[REPEAT_COUNT]:= rand_var;
    REPEAT_COUNT:= REPEAT_COUNT+1;
	if (REPEAT_COUNT > REPEAT_MAX) then
		begin
		REPEAT_TABLE_FULL:= TRUE;
		REPEAT_COUNT:= 1;
		end;
	end;
end;


(*------------------------------*)
function calc_16bit_rand:int_type;
var	R,I:int_type;
begin
R:=0;
I:=0;
while (I < LCG_table_max) do
	begin
    with LCG_table[I] do
		begin
		SEED:= (((SEED * MULTIPLIER) + CONSTANT) mod MODULUS);
		R:= R xor (SEED >> SHIFT);
		end;
	inc(I);
	end;
Result:= (R and 65535);
end;


(*------------------------------*)
procedure recalc_seeds;
var
FINISHED		:boolean = FALSE;
begin
LCG_table_i:= 0;
REPEAT_TABLE_FULL:= FALSE;
SEEDS_EXHAUSTED:= FALSE;
FINISHED:= FALSE;
while	not SEEDS_EXHAUSTED
and		not FINISHED
do
	begin
	Inc(LCG_current_seed[LCG_table_i]);
	if	(LCG_current_seed[LCG_table_i] < LCG_table[LCG_table_i].MODULUS)
	then FINISHED:= TRUE
	else
		begin
		LCG_current_seed[LCG_table_i]:= 1;
		Inc(LCG_table_i);
		if (LCG_table_i >= LCG_table_max) then
			begin
			SEEDS_EXHAUSTED:= TRUE;
			if	(TEST_TYPE = T_OP_WRITE_SEEDS) then
				writeln('SEEDS_EXHAUSTED');
			end;
		end;
	end;

if	(TEST_TYPE = T_OP_WRITE_SEEDS) then
	write('LCG_current_seeds =');
LCG_table_i:= 0;
FINISHED:= FALSE;
while (LCG_table_i < LCG_table_max) do
	begin
	LCG_table[LCG_table_i].SEED:= LCG_current_seed[LCG_table_i];
	if	(TEST_TYPE = T_OP_WRITE_SEEDS) then
    	write(' ',LCG_current_seed[LCG_table_i]);
	Inc(LCG_table_i);
	end;
if	(TEST_TYPE = T_OP_WRITE_SEEDS) then writeln;

end;


(*------------------------------*)
function get_16bit_rand:int_type;
label START;
var
R:int_type;
T:boolean;
begin
START:
while NOT REPEAT_TABLE_FULL do
	begin
	R:= calc_16bit_rand;
	T:= rand_repeat_test(R);
	end;
R:= calc_16bit_rand;
T:= rand_repeat_test(R);
if T then
	begin
	recalc_seeds;
	goto START;
	end;
Result:= R;
end;


(*--------------v2--------------*)
function prime_rand:int_type;
var	R,S	:int_type;

begin
if (RANDNUM_PARAM >= 0) then
	Result:= (RANDNUM_PARAM mod RANDMODNUM)
else
	begin
	R:= get_16bit_rand;

	if	(OPERATION = T_OP_WRITE_RNG)
	and	(TEST_TYPE = T_WRITE_RNG_TEXT)
	and	(	(WRITE_RNG_BITSIZE = 32)
		or	(WRITE_RNG_BITSIZE = 24)
		)
	then
		begin
		S:= get_16bit_rand;
		R:= (S or ( R << 16 ));
		end;

	Result:= R mod RANDMODNUM;
	end;
end;


(*------------------------------*)
procedure read_test_params;
var
i	:int_type;
begin
TEST_MODE:= FALSE;
OPERATION:= T_OP_NULL;
MILLISECS_PARAM:= 0;

i:= 1;
while (i <= ParamCount) do
	begin
    if	( ParamStr(i) = '-t') then
    	begin
		TEST_MODE:= TRUE;
		inc(i);
		if (i <= ParamCount) then
			begin
            if (ParamStr(i) = 'WRITE_RNG_BIN') then
				begin
				OPERATION:= T_OP_WRITE_RNG;
				TEST_TYPE:= T_WRITE_RNG_BIN;
				RANDMODNUM:= INT16U_MAXINT_1;
				inc(i);
	            end
            else if (ParamStr(i) = 'WRITE_PRIMES') then
				begin
				OPERATION:= T_OP_WRITE_PRIMES;
				inc(i);
	            end
            else if (ParamStr(i) = 'WRITE_INDEXES') then
				begin
				OPERATION:= T_OP_WRITE_INDEXES;
				inc(i);
	            end
            else if (ParamStr(i) = 'WRITE_SEEDS') then
				begin
				OPERATION:= T_OP_WRITE_RNG;
				TEST_TYPE:= T_OP_WRITE_SEEDS;
				inc(i);
				if (i <= ParamCount) then
					begin
					WRITE_RNG_BITSIZE:= strtoint64def(ParamStr(i),0);
					if	(WRITE_RNG_BITSIZE = 8) then
						RANDMODNUM:= INT8U_MAXINT_1
					else if	(WRITE_RNG_BITSIZE = 16) then
						RANDMODNUM:= INT16U_MAXINT_1
					else if	(WRITE_RNG_BITSIZE = 24) then
						RANDMODNUM:= 16777216  //  2^24
					else if	(WRITE_RNG_BITSIZE = 32) then
						RANDMODNUM:= INT32U_MAXINT_1
					;
					inc(i);
		            end
	            end
            else if (ParamStr(i) = 'WRITE_RNG_TEXT') then
				begin
				OPERATION:= T_OP_WRITE_RNG;
				TEST_TYPE:= T_WRITE_RNG_TEXT;
				inc(i);
				if (i <= ParamCount) then
					begin
					WRITE_RNG_BITSIZE:= strtoint64def(ParamStr(i),0);
					if	(WRITE_RNG_BITSIZE = 8) then
						RANDMODNUM:= INT8U_MAXINT_1
					else if	(WRITE_RNG_BITSIZE = 16) then
						RANDMODNUM:= INT16U_MAXINT_1
					else if	(WRITE_RNG_BITSIZE = 24) then
						RANDMODNUM:= 16777216  //  2^24
					else if	(WRITE_RNG_BITSIZE = 32) then
						RANDMODNUM:= INT32U_MAXINT_1
					;
					inc(i);
		            end
	            end
            else if (ParamStr(i) = 'NULL') then
				begin
				inc(i);
	            end;
            end
        end
	else if	( ParamStr(i) = '-n') then
    	begin
		inc(i);
		if (i <= ParamCount) then
			begin
            NUMBER_PARAM:= strtoint64def(ParamStr(i),1000000);
			inc(i);
            end;
		end
	else if	( ParamStr(i) = '-m') then
    	begin
		inc(i);
		if (i <= ParamCount) then
			begin
            MILLISECS_PARAM:= strtoint64def(ParamStr(i),0);
			inc(i);
            end;
		end
	else if	( ParamStr(i) = '-r') then
    	begin
		inc(i);
		if (i <= ParamCount) then
			begin
            RANDNUM_PARAM:= StrToInt64Def(ParamStr(i),-1);
			inc(i);
            end;
		end
	else inc(i);
	end;

if	OPERATION = T_OP_NULL then
	begin
	ERROR_CODE:= T_ERR_FILE;
	ERROR_MESS:= 'Operation must be specified';
	exit;
	end;

end;


(*------------------------------*)
procedure read_params;
begin
if (ParamCount=0) then
	begin
	ERROR_CODE:= T_ERR_PARAM;
	ERROR_MESS
		:=  WHO_I_AM_VERSION
		+	sLineBreak
		+	'EncPRNG -t WRITE_RNG_TEXT <8|16|24> -n <number-of-rands>'
		+	sLineBreak
		+	'EncPRNG -t WRITE_RNG_BIN -n <number-of-rands>'
		+	sLineBreak
		+	'Example: EncPRNG -t WRITE_RNG_BIN -n 1000'
		+	sLineBreak
		+	'Example: EncPRNG -t WRITE_RNG_TEXT 32 -n 1000'
		+	sLineBreak
		+	'Example: EncPRNG -t WRITE_PRIMES -n 1000 -m <daysince>'
		+	sLineBreak
		+	'Example: EncPRNG -t WRITE_INDEXES -n 1000 -m <daysince>'
		;
	exit;
	end;

read_test_params;
end;


(*------------------------------*)
procedure write_rng_text(num:int_type);
var
i,r	:int_type;
s	:string_type;
begin
i:= 0;
while (i < num) do
	begin
	r:= prime_rand;
	if (WRITE_RNG_BITSIZE = 32) then
		s:= format('%10d',[r])
	else if (WRITE_RNG_BITSIZE = 24) then
		s:= format('%8d',[r])
	else if (WRITE_RNG_BITSIZE = 16) then
		s:= format('%5d',[r])
	else if (WRITE_RNG_BITSIZE = 8) then
		s:= format('%3d',[r])
	else
		begin
		writeln('WRITE_RNG_BITSIZE must be 8, 16, 24 or 32');
		halt;
		end
	;
	writeln(s);
	inc(i);
    end;
end;


(*------------------------------*)
procedure write_rng_bin(num:int_type);
var
i,r	:int_type;
b	:int8u;

begin
i:= 0;
while (i < num) do
	begin
	r:= prime_rand;
	b:= (r mod INT8U_MAXINT_1);
	write(chr(b));
	inc(i);
	b:= (r div INT8U_MAXINT_1);
	write(chr(b));
	inc(i);
    end;
end;


(*------------------------------*)
procedure write_rng_silent(num:int_type);
var
i,r	:int_type;
begin
i:= 0;
while (i < num) do
	begin
	r:= prime_rand;
	inc(i);
    end;
end;


(*------------------------------*)
procedure write_rng;
begin
if	(TEST_TYPE = T_WRITE_RNG_BIN) then
	write_rng_bin(NUMBER_PARAM)
else if	(TEST_TYPE = T_WRITE_RNG_TEXT) then
	write_rng_text(NUMBER_PARAM)
else if	(TEST_TYPE = T_OP_WRITE_SEEDS) then
	write_rng_silent(NUMBER_PARAM)
;
end;


(*--------------v2--------------*)
procedure write_primes;
var	I:int_type = 0;
begin
while (I < LCG_table_max) do
	begin
    with LCG_table[I] do
		begin
		write(LCG_table[I].MULTIPLIER,' ');
		write(LCG_table[I].MODULUS,' ');
		write(LCG_table[I].SHIFT,' ');
		writeln;
		end;
	inc(I);
	end;
end;


(*------------------------------*)
procedure init_pass_rand_TEST;
begin
REPEAT_TABLE_FULL:= FALSE;

LCG_table_max:= 3;
setlength(LCG_table,LCG_table_max);
setlength(LCG_current_seed,LCG_table_max);

LCG_table[0].MULTIPLIER:= 111;
LCG_table[0].MODULUS:= 103;
LCG_table[0].CONSTANT:= 1;
LCG_table[0].SHIFT:= 0;
LCG_table[0].SEED:= 1;
LCG_current_seed[0]:= 1;

LCG_table[1].MULTIPLIER:= 107;
LCG_table[1].MODULUS:= 101;
LCG_table[1].CONSTANT:= 1;
LCG_table[1].SHIFT:= 0;
LCG_table[1].SEED:= 1;
LCG_current_seed[1]:= 1;

LCG_table[2].MULTIPLIER:= 113;
LCG_table[2].MODULUS:= 109;
LCG_table[2].CONSTANT:= 1;
LCG_table[2].SHIFT:= 0;
LCG_table[2].SEED:= 1;
LCG_current_seed[2]:= 1;

if	(TEST_TYPE = T_OP_WRITE_SEEDS) then
	begin
	LCG_table_i:= 0;
	write('LCG_current_seeds =');
	while (LCG_table_i < LCG_table_max) do
		begin
       	write(' ',LCG_current_seed[LCG_table_i]);
		Inc(LCG_table_i);
		end;
	writeln;
	end;
end;


(*------------------------------*)
procedure initialise_Encrypt_P;
begin

{$IFDEF 64bit}
Multi_Init_Initialisation(48);
{$ENDIF}

{$IFDEF 32bit}
Multi_Init_Initialisation(96);
{$ENDIF}

init_pass_rand;
// init_pass_rand_TEST;
if	(ERROR_CODE <> T_ERR_OK) then exit;
end;


end.

